@@ -1,3 +1,4 @@
+/* -*- mode: C; c-file-style: "linux" -*- */
 /* GdkPixbuf library - SUNRAS image loader
  *
  * Copyright (C) 1999 The Free Software Foundation
@@ -30,7 +31,7 @@ Known bugs:
 
 */
 
-#include <config.h>
+#include "config.h"
 #include <stdio.h>
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
@@ -67,8 +68,9 @@ struct rasterfile {
 /* Progressive loading */
 
 struct ras_progressive_state {
-	ModulePreparedNotifyFunc prepared_func;
-	ModuleUpdatedNotifyFunc updated_func;
+	GdkPixbufModuleSizeFunc size_func;
+	GdkPixbufModulePreparedFunc prepared_func;
+	GdkPixbufModuleUpdatedFunc updated_func;
 	gpointer user_data;
 
 	gint HeaderSize;	/* The size of the header-part (incl colormap) */
@@ -85,7 +87,7 @@ struct ras_progressive_state {
 				   8 = 8 bit colormapped
 				   1  = 1 bit bitonal 
 				 */
-
+	gint DecoderState;        
 
 	struct rasterfile Header;	/* Decoded (BE->CPU) header */
 
@@ -94,8 +96,9 @@ struct ras_progressive_state {
 };
 
 static gpointer
-gdk_pixbuf__ras_image_begin_load(ModulePreparedNotifyFunc prepared_func,
-				 ModuleUpdatedNotifyFunc updated_func,
+gdk_pixbuf__ras_image_begin_load(GdkPixbufModuleSizeFunc size_func,
+                                 GdkPixbufModulePreparedFunc prepared_func,
+				 GdkPixbufModuleUpdatedFunc updated_func,
 				 gpointer user_data,
                                  GError **error);
 static gboolean gdk_pixbuf__ras_image_stop_load(gpointer data, GError **error);
@@ -103,43 +106,9 @@ static gboolean gdk_pixbuf__ras_image_load_increment(gpointer data,
                                                      const guchar * buf, guint size,
                                                      GError **error);
 
-
-
-/* Shared library entry point */
-static GdkPixbuf *gdk_pixbuf__ras_image_load(FILE * f, GError **error)
-{
-	guchar *membuf;
-	size_t length;
-	struct ras_progressive_state *State;
-	
-	GdkPixbuf *pb;
-	
-	State = gdk_pixbuf__ras_image_begin_load(NULL, NULL, NULL, error);
-	
-	membuf = g_malloc(4096);
-	
-	g_assert(membuf != NULL);
-	
-	while (feof(f) == 0) {
-		length = fread(membuf, 1, 4096, f);
-                if (!gdk_pixbuf__ras_image_load_increment(State, membuf, length,
-                                                          error)) {
-                        gdk_pixbuf__ras_image_stop_load (State, NULL);
-                        return NULL;
-                }
-	}
-	g_free(membuf);
-	if (State->pixbuf != NULL)
-		g_object_ref(State->pixbuf);
-
-	pb = State->pixbuf;
-
-	gdk_pixbuf__ras_image_stop_load(State, NULL);
-	return pb;
-}
-
-static void RAS2State(struct rasterfile *RAS,
-		      struct ras_progressive_state *State)
+static gboolean RAS2State(struct rasterfile *RAS,
+			  struct ras_progressive_state *State,
+			  GError **error)
 {
 	State->Header.width = GUINT32_FROM_BE(RAS->width);
 	State->Header.height = GUINT32_FROM_BE(RAS->height);
@@ -148,47 +117,90 @@ static void RAS2State(struct rasterfile *RAS,
 	State->Header.maptype = GUINT32_FROM_BE(RAS->maptype);
 	State->Header.maplength = GUINT32_FROM_BE(RAS->maplength);
 
-	g_assert(State->Header.maplength <= 768);	/* Otherwise, we are in trouble */
+	if ((gint)State->Header.width <= 0 ||
+	    (gint)State->Header.height <= 0 || 
+	    State->Header.maplength > 768) {
+		g_set_error_literal (error,
+                                     GDK_PIXBUF_ERROR,
+                                     GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
+                                     _("RAS image has bogus header data"));
+		return FALSE;
+	}
 
 	State->RasType = State->Header.depth;	/* This may be less trivial someday */
 	State->HeaderSize = 32 + State->Header.maplength;
 
 	if (State->RasType == 32)
 		State->LineWidth = State->Header.width * 4;
-	if (State->RasType == 24)
+	else if (State->RasType == 24)
 		State->LineWidth = State->Header.width * 3;
-	if (State->RasType == 8)
+	else if (State->RasType == 8)
 		State->LineWidth = State->Header.width * 1;
-	if (State->RasType == 1) {
+	else if (State->RasType == 1) {
 		State->LineWidth = State->Header.width / 8;
 		if ((State->Header.width & 7) != 0)
 			State->LineWidth++;
 	}
+	else {
+		g_set_error_literal (error,
+                                     GDK_PIXBUF_ERROR,
+                                     GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
+                                     _("RAS image has unknown type"));
+		return FALSE;
+	}
 
-	/* Now padd the line to be a multiple of 16 bits */
+	if (State->Header.type > 2 || State->Header.maptype > 1) {
+		g_set_error_literal (error,
+                                     GDK_PIXBUF_ERROR,
+                                     GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
+                                     _("unsupported RAS image variation"));
+		return FALSE;
+	}
+
+	/* Now pad the line to be a multiple of 16 bits */
 	if ((State->LineWidth & 1) != 0)
 		State->LineWidth++;
 
-	if (State->LineBuf == NULL)
-		State->LineBuf = g_malloc(State->LineWidth);
+	if (!State->LineBuf) {
+		State->LineBuf = g_try_malloc (State->LineWidth);
 
-	g_assert(State->LineBuf != NULL);
+		if (!State->LineBuf) {
+			g_set_error_literal (error,
+                                             GDK_PIXBUF_ERROR,
+                                             GDK_PIXBUF_ERROR_INSUFFICIENT_MEMORY,
+                                             _("Not enough memory to load RAS image"));
+			return FALSE;
+		}
+	}
 
 
-	if (State->pixbuf == NULL) {
+	if (!State->pixbuf) {
+		if (State->size_func) {
+			gint width = State->Header.width;
+			gint height = State->Header.height;
+			
+			(*State->size_func) (&width, &height, State->user_data);
+			if (width == 0 || height == 0)
+				return FALSE;
+		}
+
 		if (State->RasType == 32)
-			State->pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, TRUE,
-						       8,
-						       (gint)
-						       State->Header.width,
-						       (gint)
-						       State->Header.
-						       height);
+			State->pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8,
+							(gint) State->Header.width,
+							(gint) State->Header.height);
 		else
-			State->pixbuf =
-			    gdk_pixbuf_new(GDK_COLORSPACE_RGB, FALSE, 8,
-					   (gint) State->Header.width,
-					   (gint) State->Header.height);
+			State->pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8,
+							(gint) State->Header.width,
+							(gint) State->Header.height);
+		
+                if (!State->pixbuf) {
+                        g_set_error_literal (error,
+                                             GDK_PIXBUF_ERROR,
+                                             GDK_PIXBUF_ERROR_INSUFFICIENT_MEMORY,
+                                             _("Not enough memory to load RAS image"));
+                        return FALSE;
+                }
+
 		if (State->prepared_func != NULL)
 			/* Notify the client that we are ready to go */
 			(*State->prepared_func) (State->pixbuf,
@@ -197,7 +209,7 @@ static void RAS2State(struct rasterfile *RAS,
 
 	}
 	
-	if ((State->Header.maplength==0)&&(State->RasType==1)) {
+	if ((State->Header.maplength == 0) && (State->RasType == 1)) {
 		State->HeaderBuf[32] = 255;
 		State->HeaderBuf[33] = 0;
 		State->HeaderBuf[34] = 255;
@@ -206,6 +218,7 @@ static void RAS2State(struct rasterfile *RAS,
 		State->HeaderBuf[37] = 0;
 	}
 
+	return TRUE;
 }
 
 /* 
@@ -215,14 +228,16 @@ static void RAS2State(struct rasterfile *RAS,
  */
 
 static gpointer
-gdk_pixbuf__ras_image_begin_load(ModulePreparedNotifyFunc prepared_func,
-				 ModuleUpdatedNotifyFunc updated_func,
+gdk_pixbuf__ras_image_begin_load(GdkPixbufModuleSizeFunc size_func, 
+                                 GdkPixbufModulePreparedFunc prepared_func,
+				 GdkPixbufModuleUpdatedFunc updated_func,
 				 gpointer user_data,
                                  GError **error)
 {
 	struct ras_progressive_state *context;
 
 	context = g_new0(struct ras_progressive_state, 1);
+	context->size_func = size_func;
 	context->prepared_func = prepared_func;
 	context->updated_func = updated_func;
 	context->user_data = user_data;
@@ -238,7 +253,8 @@ gdk_pixbuf__ras_image_begin_load(ModulePreparedNotifyFunc prepared_func,
 	context->Lines = 0;
 
 	context->RasType = 0;
-
+	context->DecoderState = 0;
+	
 	memset(&context->Header, 0, sizeof(struct rasterfile));
 
 
@@ -265,10 +281,8 @@ gdk_pixbuf__ras_image_stop_load(gpointer data, GError **error)
 
 	g_return_val_if_fail(context != NULL, TRUE);
 
-	if (context->LineBuf != NULL)
-		g_free(context->LineBuf);
-	if (context->HeaderBuf != NULL)
-		g_free(context->HeaderBuf);
+	g_free(context->LineBuf);
+	g_free(context->HeaderBuf);
 
 	if (context->pixbuf)
 		g_object_unref(context->pixbuf);
@@ -321,6 +335,7 @@ static void OneLine8(struct ras_progressive_state *context)
 {
 	gint X;
 	guchar *Pixels;
+	int offset = context->Header.maplength / 3;
 
 	X = 0;
 	Pixels = context->pixbuf->pixels + context->pixbuf->rowstride * context->Lines;
@@ -329,9 +344,9 @@ static void OneLine8(struct ras_progressive_state *context)
 		Pixels[X * 3 + 0] =
 		    context->HeaderBuf[context->LineBuf[X] + 32];
 		Pixels[X * 3 + 1] =
-		    context->HeaderBuf[context->LineBuf[X] + 256 + 32];
+		    context->HeaderBuf[context->LineBuf[X] + offset + 32];
 		Pixels[X * 3 + 2] =
-		    context->HeaderBuf[context->LineBuf[X] + 512 + 32];
+		    context->HeaderBuf[context->LineBuf[X] + 2*offset + 32];
 		X++;
 	}
 }
@@ -362,6 +377,9 @@ static void OneLine1(struct ras_progressive_state *context)
 
 static void OneLine(struct ras_progressive_state *context)
 {
+	context->LineDone = 0;
+	if (context->Lines >= context->Header.height)
+		return;
 	if (context->RasType == 32)
 		OneLine32(context);
 	if (context->RasType == 24)
@@ -372,8 +390,6 @@ static void OneLine(struct ras_progressive_state *context)
 		OneLine1(context);
 
 	context->LineDone = 0;
-	if (context->Lines > context->Header.height)
-		return;
 	context->Lines++;
 
 	if (context->updated_func != NULL) {
@@ -381,18 +397,54 @@ static void OneLine(struct ras_progressive_state *context)
 					  0,
 					  context->Lines,
 					  context->Header.width,
-					  context->Header.height,
+					  1,
 					  context->user_data);
 
 	}
 }
 
+static gboolean
+DoCompressed (struct ras_progressive_state *context,
+	      const guchar * buf, guint size,
+	      GError **error)
+{
+	int i;
+
+	for (i = 0; i < size; i++) {
+		switch (context->DecoderState) {
+		    case 0:
+			    if (buf[i] == 0x80)
+				    context->DecoderState = 1;
+			    else
+				    context->LineBuf[context->LineDone++] = buf[i];
+			    break;
+		    case 1:
+			    if (buf[i] == 0) {
+				    context->LineBuf[context->LineDone++] = 0x80;
+				    context->DecoderState = 0;
+			    }
+			    else
+				    context->DecoderState = buf[i] + 1;
+			    break;
+		    default:
+			    for (; context->DecoderState; context->DecoderState--) {
+				    context->LineBuf[context->LineDone++] = buf[i];
+				    if ((context->LineDone >= context->LineWidth) && (context->LineWidth > 0))
+					    OneLine(context);
+			    }
+		}
+		if ((context->LineDone >= context->LineWidth) && (context->LineWidth > 0))
+			OneLine(context);
+	}
+	return TRUE;
+}
+
 /*
  * context - from image_begin_load
  * buf - new image data
  * size - length of new image data
  *
- * append image data onto inrecrementally built output image
+ * append image data onto incrementally built output image
  */
 static gboolean
 gdk_pixbuf__ras_image_load_increment(gpointer data,
@@ -419,8 +471,13 @@ gdk_pixbuf__ras_image_load_increment(gpointer data,
 			buf += BytesToCopy;
 			context->HeaderDone += BytesToCopy;
 
-		} else {
-			/* Pixeldata only */
+		} else if (context->Header.type == 2) {
+			if (!DoCompressed (context, buf, size, error)) {
+				return FALSE;
+			}
+			size = 0;
+		}
+		else {
 			BytesToCopy =
 			    context->LineWidth - context->LineDone;
 			if (BytesToCopy > size)
@@ -443,8 +500,10 @@ gdk_pixbuf__ras_image_load_increment(gpointer data,
 		}
 
 		if (context->HeaderDone >= 32)
-			RAS2State((struct rasterfile *) context->HeaderBuf,
-				  context);
+			if (!RAS2State((struct rasterfile *) context->HeaderBuf,
+				       context, error)) {
+				return FALSE;
+			}
 
 
 	}
@@ -452,11 +511,41 @@ gdk_pixbuf__ras_image_load_increment(gpointer data,
 	return TRUE;
 }
 
-void
-gdk_pixbuf__ras_fill_vtable (GdkPixbufModule *module)
+#ifndef INCLUDE_ras
+#define MODULE_ENTRY(function) G_MODULE_EXPORT void function
+#else
+#define MODULE_ENTRY(function) void _gdk_pixbuf__ras_ ## function
+#endif
+
+MODULE_ENTRY (fill_vtable) (GdkPixbufModule *module)
 {
-  module->load = gdk_pixbuf__ras_image_load;
-  module->begin_load = gdk_pixbuf__ras_image_begin_load;
-  module->stop_load = gdk_pixbuf__ras_image_stop_load;
-  module->load_increment = gdk_pixbuf__ras_image_load_increment;
+	module->begin_load = gdk_pixbuf__ras_image_begin_load;
+	module->stop_load = gdk_pixbuf__ras_image_stop_load;
+	module->load_increment = gdk_pixbuf__ras_image_load_increment;
 }
+
+MODULE_ENTRY (fill_info) (GdkPixbufFormat *info)
+{
+	static GdkPixbufModulePattern signature[] = {
+		{ "\x59\xa6\x6a\x95", NULL, 100 },
+		{ NULL, NULL, 0 }
+	};
+	static gchar * mime_types[] = {
+		"image/x-cmu-raster",
+		"image/x-sun-raster",
+		NULL
+	};
+	static gchar * extensions[] = {
+		"ras",
+		NULL
+	};
+
+	info->name = "ras";
+	info->signature = signature;
+	info->description = N_("The Sun raster image format");
+	info->mime_types = mime_types;
+	info->extensions = extensions;
+	info->flags = GDK_PIXBUF_FORMAT_THREADSAFE;
+	info->license = "LGPL";
+}
+
