@@ -24,12 +24,12 @@
  * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
  */
 
-#include <config.h>
-#include "gdkalias.h"
+#include "config.h"
 #include "gdkpixmap.h"
 #include "gdkinternals.h"
 #include "gdkpixbuf.h"
 #include "gdkscreen.h"
+#include "gdkalias.h"
 
 static GdkGC *gdk_pixmap_create_gc      (GdkDrawable     *drawable,
                                          GdkGCValues     *values,
@@ -144,6 +144,8 @@ static GdkImage* gdk_pixmap_copy_to_image (GdkDrawable *drawable,
 					   gint         width,
 					   gint         height);
 
+static cairo_surface_t *gdk_pixmap_ref_cairo_surface (GdkDrawable *drawable);
+
 static GdkVisual*   gdk_pixmap_real_get_visual   (GdkDrawable *drawable);
 static gint         gdk_pixmap_real_get_depth    (GdkDrawable *drawable);
 static void         gdk_pixmap_real_set_colormap (GdkDrawable *drawable,
@@ -163,24 +165,13 @@ gdk_pixmap_get_type (void)
   static GType object_type = 0;
 
   if (!object_type)
-    {
-      static const GTypeInfo object_info =
-      {
-        sizeof (GdkPixmapObjectClass),
-        (GBaseInitFunc) NULL,
-        (GBaseFinalizeFunc) NULL,
-        (GClassInitFunc) gdk_pixmap_class_init,
-        NULL,           /* class_finalize */
-        NULL,           /* class_data */
-        sizeof (GdkPixmapObject),
-        0,              /* n_preallocs */
-        (GInstanceInitFunc) gdk_pixmap_init,
-      };
-      
-      object_type = g_type_register_static (GDK_TYPE_DRAWABLE,
-                                            "GdkPixmap",
-                                            &object_info, 0);
-    }
+    object_type = g_type_register_static_simple (GDK_TYPE_DRAWABLE,
+						 "GdkPixmap",
+						 sizeof (GdkPixmapObjectClass),
+						 (GClassInitFunc) gdk_pixmap_class_init,
+						 sizeof (GdkPixmapObject),
+						 (GInstanceInitFunc) gdk_pixmap_init,
+						 0);
   
   return object_type;
 }
@@ -224,6 +215,7 @@ gdk_pixmap_class_init (GdkPixmapObjectClass *klass)
   drawable_class->get_colormap = gdk_pixmap_real_get_colormap;
   drawable_class->get_visual = gdk_pixmap_real_get_visual;
   drawable_class->_copy_to_image = gdk_pixmap_copy_to_image;
+  drawable_class->ref_cairo_surface = gdk_pixmap_ref_cairo_surface;
 }
 
 static void
@@ -517,6 +509,12 @@ gdk_pixmap_copy_to_image (GdkDrawable     *drawable,
 				     width, height);
 }
 
+static cairo_surface_t *
+gdk_pixmap_ref_cairo_surface (GdkDrawable *drawable)
+{
+  return _gdk_drawable_ref_cairo_surface (((GdkPixmapObject*)drawable)->impl);
+}
+
 static GdkBitmap *
 make_solid_mask (GdkScreen *screen, gint width, gint height)
 {
@@ -671,3 +669,6 @@ gdk_pixmap_real_get_screen (GdkDrawable *drawable)
 {
     return gdk_drawable_get_screen (GDK_PIXMAP_OBJECT (drawable)->impl);
 }
+
+#define __GDK_PIXMAP_C__
+#include "gdkaliasdef.c"
